import DataChannel, {PeerDataChannel, SocketDataChannel} from "./DataChannel.js";

export const baseURL = `ws://${document.location.hostname}:443/skribbl`;
/**
 * @typedef {{connection:RTCPeerConnection,dataChannel:RTCDataChannel}} ConnectionData 
 */
/**
 * A connection to the signaling server used to start and join games.
 */
export default class Signaler {

	/**
	 * Starts a new game, returning a promise with its ID. After that, calls the given callback with a new RTCPeerConnection whenever someone joins the game.
	 * @param {(dataChannel:PeerDataChannel)=>void} onConnect
	 */
	static async host(onConnect){
		const socket = new SocketDataChannel(baseURL+"/host");
		//await socket.waitUntilReady();
		let response = await socket.next();
		console.log("response:",response);
		if (!response.startsWith("started ")){
			throw new Error(`Unexpected first response or something: "${response}"`);
		}
		let id = response.split(" ")[1];
		socket.onMessage(async message=>{
			if (message.startsWith("join ")){
				let playerID = message.split(" ",2)[1];
				console.log("new player joined: ",playerID);
				let signalingChannel = DataChannel.from(socket,message=>`message ${playerID} ${message}`,message=>{
					if (message.startsWith("message "+playerID+" ")){
						return message.split(" ").slice(2).join(" ");
					}
				});
				onConnect(await perfectNegotiation(signalingChannel,false));
			}
		});
		return id;
	}

	/**
	 * Joins a new game, returning an RTCPeerConnection to its host.
	 * @param {string} id
	 */
	static async join(id){
		const socket = new SocketDataChannel(baseURL+"/join/"+id);
		let signalingChannel = DataChannel.from(socket,message=>"message "+message,message=>{
			if (message.startsWith("message ")){
				return message.split(" ").slice(1).join(" ");
			}else{
				throw new Error(`Unexpected message: "${message}"`);
			}
		});
		return await perfectNegotiation(signalingChannel,true);
	}
}
/** 
 * An implementation of "perfect negotiaton" as described [here](https://developer.mozilla.org/en-US/docs/Web/API/WebRTC_API/Perfect_negotiation).
 * @param {DataChannel<string,string>} signalingChannel
 * @param {boolean} polite
 * @returns {Promise<PeerDataChannel>}
 * @todo figure out connection issues that sometimes randomly occur
 */
export async function perfectNegotiation(signalingChannel,polite){
	await signalingChannel.waitUntilReady();
	const connection = new RTCPeerConnection();
	const dataChannel = connection.createDataChannel("channel",{id:0,negotiated:true});
	let makingOffer = false;
	connection.onnegotiationneeded = async()=>{
		makingOffer = true;
		// @ts-ignore
		await connection.setLocalDescription();
		//console.log("local description:",connection.localDescription)
		signalingChannel.send(JSON.stringify({description:connection.localDescription}));
		makingOffer = false;
	};
	connection.onicecandidate = ({candidate})=>{
		//console.log("candidate:",candidate);
		signalingChannel.send(JSON.stringify({candidate}));
	};
	let ignoreOffer = false;
	signalingChannel.onMessage(async message=>{
		console.log("message to negotiator:",message);
		/** @type {{description:RTCSessionDescriptionInit,candidate:undefined}|{candidate:RTCIceCandidateInit,description:undefined}} */
		let data = JSON.parse(message);
		if (data.description){
			let description = data.description;
			ignoreOffer = !polite&&(description.type=="offer"&&(makingOffer||connection.signalingState!="stable"));
			if(!ignoreOffer){
				await connection.setRemoteDescription(description);
				if (description.type=="offer"){
					// @ts-ignore
					await connection.setLocalDescription();
					signalingChannel.send(JSON.stringify({description:connection.localDescription}));
				}
			}
		}else if(data.candidate){
			try {
				await connection.addIceCandidate(data.candidate);
			}catch(e){
				if (!ignoreOffer){
					throw e;
				}
			}
		}
	});
	return new PeerDataChannel(connection,dataChannel);
}